--=====================================================================================
-- SELF PROPELLED WINDROWER
--
-- @Author:		akuenzi
-- @Date:		2018-1-31
-- @Purpose:	Allows mower to drop a dryGrass windrow instead of a grass windrow.
--=====================================================================================


if SPW == nil then
	-- Hour of day (0-23) when cutting for grass seed can start/end without consequence of shattering seed.
	SPW = {}
	SPW.GRASSEED_CUT_ACTIVATE	=	false		-- Activate/Deactivate this part of the script (true/false)
	SPW.CURRENT_SHATTER_FACTOR	=	1			-- The active shatter factor based on time.

	--[[Following is a table to show shattering losses, which will basically manifest as a reduction of the windrow size.
		The number on the right (commented out) is the hour of the day.  The number on the left is the portion of seed kept
		if cutting grass seed at that hour.  A factor of 1 means there is no shattering loss.  A factor of 0.7 means the 
		resultant windrow is 70% of what it should be (ie, 30% shattered).]]
	SPW.SHATTER = {	
					1,		--[[12:00am]]
					1,		--[[1:00am]]
					1,		--[[2:00am]]
					1,		--[[3:00am]]
					1,		--[[4:00am]]
					1,		--[[5:00am]]
					1,		--[[6:00am]]
					1,		--[[7:00am]]
					0.95,	--[[8:00am]]
					0.9,	--[[9:00am]]
					0.85,	--[[10:00am]]
					0.8,	--[[11:00am]]
					0.75,	--[[12:00pm]]
					0.7,	--[[1:00pm]]
					0.65,	--[[2:00pm]]
					0.6,	--[[3:00pm]]
					0.5,	--[[4:00pm]]
					0.5,	--[[5:00pm]]
					0.65,	--[[6:00pm]]
					0.8,	--[[7:00pm]]
					1,		--[[8:00pm]]
					1,		--[[9:00pm]]
					1,		--[[10:00pm]]
					1,		--[[11:00pm]]
					}
end;


SelfPropelledWindrower = {};				
				
function SelfPropelledWindrower.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Mower, specializations)
end;

function SelfPropelledWindrower:load(savegame)
	self.SPWToggle = false																-- On/off toggle for grass seed windrow.
	self.processMowerAreas = SelfPropelledWindrower.processMowerAreas					-- To replace existing GIANTS script for processing mower areas for this vehicle.
	self.setSPWStatus = SpecializationUtil.callSpecializationsFunction("setSPWStatus");	-- On/off toggle status update specialization.

	-- Check to see if grass seed windrow is toggled on/off in savegame file.
	local saveToggleGrass
	if savegame ~= nil then
		saveToggleGrass = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key.."#grasseedWindrow"), false)
		self:setSPWStatus(saveToggleGrass, true)
	end;
	
end;

function SelfPropelledWindrower:delete()
end;

function SelfPropelledWindrower:readStream(streamId, connection)
	self:setSPWStatus(streamReadBool(streamId), true);
end;

function SelfPropelledWindrower:writeStream(streamId, connection)
	streamWriteBool(streamId, self.SPWToggle)
end;

function SelfPropelledWindrower:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'grasseedWindrow="' .. tostring(self.SPWToggle) .. '"';
	return attributes, nil;
end;

function SelfPropelledWindrower:setSPWStatus(toggle, noEventSend)
	if self.SPWToggle ~= toggle then
		if noEventSend == nil or noEventSend == false then
			if g_server~= nil then
				g_server:broadcastEvent(setSPWStatusEvent:new(self, toggle), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(setSPWStatusEvent:new(self, toggle));
			end;
		end;
		self.SPWToggle = toggle		-- To update on/off toggle after syncing MP.
	end;
end;

function SelfPropelledWindrower:mouseEvent(posX, posY, isDown, isUp, button)
end;

function SelfPropelledWindrower:keyEvent(unicode, sym, modifier, isDown)
end;

function SelfPropelledWindrower:update(dt)
	if self:getIsActive() then
		if self:getIsActiveForInput() then
			-- Check to see if player has enabled grass seed windrow.
			if InputBinding.hasEvent(InputBinding.GRASSEED_WINDROW) then
				local toggle
				if self.SPWToggle then
					toggle = false
				else
					toggle = true
				end;
				self:setSPWStatus(toggle)		-- Update toggle and MP syncing.
			end;
		end;
	end;
end;

function SelfPropelledWindrower:draw()
	if self.isClient then
		if self:getIsActiveForInput(true) then
			if self.SPWToggle then
				-- Grass seed windrow is currently 'on.'  Show how to turn it off.
				g_currentMission:addHelpButtonText(g_i18n:getText("input_GRASSEED_WINDROW") .. g_i18n:getText("GRASSEED_WINDROW_OFF"), InputBinding.GRASSEED_WINDROW, nil, GS_PRIO_NORMAL);
			else
				-- Grass seed windrow is currently 'off.'  Show how to turn it on.
				g_currentMission:addHelpButtonText(g_i18n:getText("input_GRASSEED_WINDROW") .. g_i18n:getText("GRASSEED_WINDROW_ON"), InputBinding.GRASSEED_WINDROW, nil, GS_PRIO_NORMAL);
			end;
		end;
		
		if self.SPWToggle and SPW.GRASSEED_CUT_ACTIVATE and SPW.CURRENT_SHATTER_FACTOR < 1 then
			local loss = (1 - SPW.CURRENT_SHATTER_FACTOR) * 100
			g_currentMission:addExtraPrintText(string.format(g_i18n:getText("GRASSEED_SHATTERING"), tostring(loss .. "%")))
		end;
	end;
end;

function SelfPropelledWindrower:processMowerAreas(workAreas, numWorkAreas, dropAreas, numDropAreas, remainingWindrowToDrop)

	local pickedUpWindrow = 0
	local numDropAreasUsed = 0
	local fillType

	-- Choose which fillType to drop if grass seed windrows are being cut.
	if self.SPWToggle then
		fillType = FruitUtil.fruitTypeToWindrowFillType[FruitUtil.FRUITTYPE_DRYGRASS]
	else
		fillType = FruitUtil.fruitTypeToWindrowFillType[FruitUtil.FRUITTYPE_GRASS]
	end;
	
	-- Now see if the windrow is cut during a time of shattering with grasseed windrows.
	if SPW ~= nil and self.SPWToggle then
		if SPW.GRASSEED_CUT_ACTIVATE then
			if g_currentMission:getIsServer() or g_currentMission.isMasterUser then
				local currentHour = g_currentMission.environment.currentHour + 1	-- Need to add 1 for proper indexing to SPW table, which starts at 1, not 0.
--				print("Current Hour: " .. tostring(currentHour - 1))
				if SPW.SHATTER[currentHour] ~= nil and SPW.SHATTER[currentHour] < 1 then
					SPW.CURRENT_SHATTER_FACTOR = SPW.SHATTER[currentHour]
				else
					SPW.CURRENT_SHATTER_FACTOR = 1
				end;
--				print("Shatter Factor:" .. tostring(SPW.CURRENT_SHATTER_FACTOR))
			end;
		end;
	else	-- If we're dropping a regular grass_windrow.
		if SPW ~= nil then
			SPW.CURRENT_SHATTER_FACTOR = 1
		end;
	end;
	
	for i=1, numWorkAreas do
		local area = workAreas[i]

		local x0 = area.x
		local z0 = area.z
		local x1 = area.x1
		local z1 = area.z1
		local x2 = area.x2
		local z2 = area.z2

		local areaPixelsSum, _, sprayFactor, _ = Utils.cutFruitArea(FruitUtil.FRUITTYPE_GRASS, x0,z0, x1,z1, x2,z2, true, true)
		if areaPixelsSum > 0 then
			local multi = g_currentMission:getHarvestScaleMultiplier(sprayFactor, 1)
			areaPixelsSum = areaPixelsSum * multi
		end

		local pixelToSqm = g_currentMission:getFruitPixelsToSqm()
		local sqm = areaPixelsSum * pixelToSqm
		
		--[[litersChanged represents amount of additional liters cut by the mower.  This gets added to the remainingWindrowToDrop.  As stuff is dropped from the windrow
			(see lines 199, 202, or 214), the amount dropped gets subtracted from this total.  If there is shattering occurring, then we want there to be LESS windrow
			available to pick up by the combine in the first place, hence usage of 'SPW.CURRENT_SHATTER_FACTOR' below as a modifier.  If this part of the script is deactivated,
			then 'SPW.CURRENT_SHATTER_FACTOR' is always 1 by default.]]
		local litersChanged = sqm * FruitUtil.getFillTypeLiterPerSqm(FruitUtil.fruitTypeToWindrowFillType[FruitUtil.FRUITTYPE_GRASS], 1) * (SPW.CURRENT_SHATTER_FACTOR or 1)

		if area.dropArea ~= nil then
			dropAreas[area.dropArea].valueAccum = dropAreas[area.dropArea].valueAccum + litersChanged
		else
			if not area.dropWindrow then
				pickedUpWindrow = pickedUpWindrow + litersChanged
			else

				local lx_2 = 0.5*(area.x2 - area.x)
				local lz_2 = 0.5*(area.z2 - area.z)
				local sx = area.x + lx_2
				local sz = area.z + lz_2
				local ex = area.x1 + lx_2
				local ez = area.z1 + lz_2
				local sy = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, sx,0,sz)
				local ey = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, ex,0,ez)

				local toDrop = remainingWindrowToDrop + litersChanged
				local dropped, lineOffset = TipUtil.tipToGroundAroundLine(self, toDrop, fillType, sx,sy,sz, ex,ey,ez, 0, nil, self.mowerLineOffset, false, nil, false)
				self.mowerLineOffset = lineOffset
				remainingWindrowToDrop = remainingWindrowToDrop + (toDrop - dropped)
			end
		end

	end
	
	local unitLength = TipUtil.densityToWorldMap

	for i=1, numDropAreas do
		local area = dropAreas[i]
		local widthX = area.x1 - area.x
		local widthZ = area.z1 - area.z
		local widthLength = Utils.vector2Length(widthX, widthZ)
		local widthLength_2 = 0.5 * widthLength

		local widthX_norm = widthX / widthLength
		local widthZ_norm = widthZ / widthLength

		local middleX = 0.5 * (area.x1 + area.x)
		local middleZ = 0.5 * (area.z1 + area.z)

		local sx = middleX + ( widthX_norm * math.max(0, widthLength_2 - unitLength) )
		local sz = middleZ + ( widthZ_norm * math.max(0, widthLength_2 - unitLength) )

		local ex = middleX - ( widthX_norm * math.max(0, widthLength_2 - unitLength) )
		local ez = middleZ - ( widthZ_norm * math.max(0, widthLength_2 - unitLength) )

		local sy = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, sx,0,sz)
		local ey = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, ex,0,ez)

		--drawDebugLine(sx,sy,sz, 1,0,0, sx,sy+2,sz, 1,0,0)
		--drawDebugLine(ex,ey,ez, 0,1,0, ex,ey+2,ez, 0,1,0)

		local toDrop = area.valueAccum
		local dropped, lineOffset = TipUtil.tipToGroundAroundLine(self, toDrop, fillType, sx,sy,sz, ex,ey,ez, 0, nil, area.mowerLineOffset, false, nil, false)
		area.mowerLineOffset = lineOffset

		area.valueAccum = area.valueAccum - dropped
		area.value = dropped

		if dropped > 0 then
			numDropAreasUsed = numDropAreasUsed + 1
		end

	end

	return numDropAreasUsed, pickedUpWindrow, remainingWindrowToDrop
end;

---------------------------------------------------------------------------------------------------------
--	SET SPW TOGGLE STATUS EVENT FOR MULTIPLAYER (IE, WHEN GRASSEED_WINDROW IS TOGGLED ON/OFF)
---------------------------------------------------------------------------------------------------------
setSPWStatusEvent = {};
setSPWStatusEvent_mt = Class(setSPWStatusEvent, Event);

InitEventClass(setSPWStatusEvent, "setSPWStatusEvent");

function setSPWStatusEvent:emptyNew()
	local self = Event:new(setSPWStatusEvent_mt);
	self.className = "setSPWStatusEvent";
	return self;
end;

function setSPWStatusEvent:new(object, status)
	local self = setSPWStatusEvent:emptyNew()
	self.object = object;
	self.SPWToggle = status;
	return self;
end;

function setSPWStatusEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.SPWToggle = streamReadBool(streamId);
	self:run(connection);
end;

function setSPWStatusEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId,self.SPWToggle);
end;

function setSPWStatusEvent:run(connection)
	self.object:setSPWStatus(self.SPWToggle, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(setSPWStatusEvent:new(self.object,self.SPWToggle),nil,connection,self.object);
	end;
end;